from collections import defaultdict
from math import log10, log
from threading import Thread
import threading
import math
import pyglet
import numpy as np
import geoplotlib.colors as colors
from geoplotlib.core import BatchPainter,FONT_NAME
from geoplotlib.utils import BoundingBox
import Queue
from inspect import isfunction
import json



class HotspotManager():


    def __init__(self):
        self.rects = []
        self.poly = []


    # adapted from:
    # http://stackoverflow.com/questions/16625507/python-checking-if-point-is-inside-a-polygon
    @staticmethod
    def point_in_poly(x, y, bbox, poly):
        left, top, right, bottom = bbox
        if x < left or x > right or y < top or y > bottom:
            return False

        n = len(poly)
        is_inside = False

        x1,y1 = poly[0]
        for i in range(n+1):
            x2,y2 = poly[i % n]
            if y > min(y1,y2):
                if y <= max(y1,y2):
                    if x <= max(x1,x2):
                        if y1 != y2:
                            xints = (y-y1)*(x2-x1)/(y2-y1)+x1
                        if x1 == x2 or x <= xints:
                            is_inside = not is_inside
            x1,y1 = x2,y2

        return is_inside


    def add_rect(self, x, y, w, h, value):
        self.rects.append(((x, y, w, h), value))


    def add_poly(self, x, y, value):
        bbox = (x.min(), y.min(), x.max(), y.max())
        self.poly.append((zip(x,y), bbox, value))


    def pick(self, mouse_x, mouse_y):
        for (x, y, w, h), value in self.rects:
            if (x <= mouse_x <= x + w) and (y <= mouse_y <= y + h):
                return value

        for points, bbox, value in self.poly:
            if HotspotManager.point_in_poly(mouse_x, mouse_y, bbox, points):
                return value

        return None


class BaseLayer():
    """
    Base class for layers
    """

    def invalidate(self, proj):
        """
        This method is called each time layers need to be redrawn, i.e. on zoom.
        Typically in this method a BatchPainter is instantiated and all the rendering is performed

        :param proj: the current Projector object
        """
        pass


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        """
        This method is called at every frame, and typically executes BatchPainter.batch_draw()

        :param proj: the current Projector object
        :param mouse_x: mouse x
        :param mouse_y: mouse y
        :param ui_manager: the current UiManager
        """
        pass


    def bbox(self):
        """
        Return the bounding box for this layer
        """
        return BoundingBox.WORLD


    def on_key_release(self, key, modifiers):
        """
        Override this method for custom handling of keystrokes

        :param key: the key that has been released
        :param modifiers: the key modifiers
        :return: True if the layer needs to call invalidate
        """
        return False


class DotDensityLayer(BaseLayer):


    def __init__(self, data, color=None, point_size=2, f_tooltip=None):
        """Create a dot density map

        :param data: data access object
        :param color: color
        :param point_size: point size
        :param f_tooltip: function to return a tooltip string for a point
        """
        self.data = data
        self.color = color
        if self.color is None:
            self.color = [255,0,0]
        self.point_size = point_size
        self.f_tooltip = f_tooltip

        self.hotspots = HotspotManager()


    def invalidate(self, proj):
        self.painter = BatchPainter()
        x, y = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])
        if self.f_tooltip:
            for i in range(0, len(x)):
                record = {k: self.data[k][i] for k in self.data.keys()}
                self.hotspots.add_rect(x[i] - self.point_size, y[i] - self.point_size,
                                       2*self.point_size, 2*self.point_size,
                                       self.f_tooltip(record))
        self.painter.set_color(self.color)
        self.painter.points(x, y, 2*self.point_size, False)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        picked = self.hotspots.pick(mouse_x, mouse_y)
        if picked:
            ui_manager.tooltip(picked)


    def bbox(self):
        return BoundingBox.from_points(lons=self.data['lon'], lats=self.data['lat'])


class HistogramLayer(BaseLayer):

    def __init__(self, data, cmap='hot', alpha=220, colorscale='sqrt', binsize=16, 
                 show_tooltip=False, scalemin=0, scalemax=None, f_group=None, show_colorbar=True):
        """Create a 2D histogram

        :param data: data access object
        :param cmap: colormap name
        :param alpha: color alpha
        :param colorscale: scaling [lin, log, sqrt]
        :param binsize: size of the hist bins
        :param show_tooltip: if True, will show the value of bins on mouseover
        :param scalemin: min value for displaying a bin
        :param scalemax: max value for a bin
        :param f_group: function to apply to samples in the same bin. Default is to count
        :param show_colorbar: show colorbar
        :return:
        """
        self.data = data
        self.cmap = colors.ColorMap(cmap, alpha=alpha)
        self.binsize = binsize
        self.show_tooltip = show_tooltip
        self.scalemin = scalemin
        self.scalemax = scalemax
        self.colorscale = colorscale
        self.f_group = f_group
        if self.f_group is None:
            self.f_group = lambda grp: len(grp)
        self.show_colorbar = show_colorbar


    def invalidate(self, proj):
        self.painter = BatchPainter()
        x, y = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])
        self.data['_xbin'] = (x / self.binsize).astype(int)
        self.data['_ybin'] = (y / self.binsize).astype(int)
        uniquevalues = set([tuple(row) for row in np.vstack([self.data['_xbin'],self.data['_ybin']]).T])
        results = {(v1,v2): self.f_group(self.data.where((self.data['_xbin'] == v1) & (self.data['_ybin'] == v2))) \
                   for v1, v2 in uniquevalues}
        del self.data['_xbin']
        del self.data['_ybin']

        self.hotspot = HotspotManager()

        if self.scalemax:
            self.vmax = self.scalemax
        else:
            self.vmax = max(results.values()) if len(results) > 0 else 0

        if self.vmax >= 1:
            for (ix, iy), value in results.items():
                if value > self.scalemin:
                    self.painter.set_color(self.cmap.to_color(value, self.vmax, self.colorscale))
                    l = self.binsize
                    rx = ix * self.binsize
                    ry = iy * self.binsize

                    self.painter.rect(rx, ry, rx+l, ry+l)
                    if self.show_tooltip:
                        self.hotspot.add_rect(rx, ry, l, l, 'Value: %d' % value)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        picked = self.hotspot.pick(mouse_x, mouse_y)
        if picked:
            ui_manager.tooltip(picked)
        if self.show_colorbar:
            ui_manager.add_colorbar(self.cmap, self.vmax, self.colorscale)


    def bbox(self):
        return BoundingBox.from_points(lons=self.data['lon'], lats=self.data['lat'])


class GraphLayer(BaseLayer):

    def __init__(self, data, src_lat, src_lon, dest_lat, dest_lon,
                 linewidth=1, alpha=220, color='hot',levels=10,
                 color_by = None, seg_scale='log'):
        """Create a graph drawing a line between each pair of (src_lat, src_lon) and (dest_lat, dest_lon)

        :param data: data access object
        :param src_lat: field name of source latitude
        :param src_lon: field name of source longitude
        :param dest_lat: field name of destination latitude
        :param dest_lon: field name of destination longitude
        :param linewidth: line width
        :param alpha: color alpha
        :param color: color or colormap
        :param levels: coloring levels
        :param color_by: attribute name for color, default using node distance
        :param seg_scale: coloring data segamentation sacle, 'log' or 'lin',
            'lin' only used if not by distance
        """
        self.data = data
        self.src_lon = src_lon
        self.src_lat = src_lat
        self.dest_lon = dest_lon
        self.dest_lat = dest_lat

        self.linewidth = linewidth
        alpha = alpha
        self.color = color
        if type(self.color) == str:
            self.cmap = colors.ColorMap(self.color, alpha, levels = levels)

        if color_by is None:
            self.color_by = 'distance'
        self.seg_scale = seg_scale


    def invalidate(self, proj):
        self.painter = BatchPainter()
        x0, y0 = proj.lonlat_to_screen(self.data[self.src_lon], self.data[self.src_lat])
        x1, y1 = proj.lonlat_to_screen(self.data[self.dest_lon], self.data[self.dest_lat])

        if type(self.color) == list:
            self.painter.set_color(self.color)
            self.painter.lines(x0, y0, x1, y1, width=self.linewidth)
        else:
            if self.color_by == 'distance':
                manhattan = np.abs(x0-x1) + np.abs(y0-y1)
                vmax = manhattan.max()
                segmentations = np.logspace(0, log10(vmax), 20)
                self.seg_scale = 'log'
            else:
                manhattan = self.data[self.color_by]
                vmax = manhattan.max()
                if self.seg_scale == 'log':
                    # value 20 maybe should be optional
                    segmentations = np.logspace(0, log10(vmax), 20)
                else:
                    # linear
                    segmentations = np.linspace(0, vmax ,20)

            for i in range(len(segmentations)-1, 1, -1):
                mask = (manhattan > segmentations[i-1]) & (manhattan <= segmentations[i])
                self.painter.set_color(self.cmap.to_color(segmentations[i], vmax, self.seg_scale))
                self.painter.lines(x0[mask], y0[mask], x1[mask], y1[mask], width=self.linewidth)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()


    def bbox(self):
        return BoundingBox.from_points(lons=np.hstack([self.data[self.src_lon], self.data[self.dest_lon]]),
                                       lats=np.hstack([self.data[self.src_lat], self.data[self.dest_lat]]))


class ShapefileLayer(BaseLayer):

    def __init__(self, fname, f_tooltip=None, color=None, linewidth=3, shape_type='full', encoding='utf-8', encodingErrors='strict'):
        """
        Loads and draws shapefiles

        :param fname: full path to the shapefile
        :param f_tooltip: function to generate a tooltip on mouseover
        :param color: color
        :param linewidth: line width
        :param shape_type: either full or bbox
        """
        if color is None:
            color = [255, 0, 0]
        self.color = color
        self.linewidth = linewidth
        self.f_tooltip = f_tooltip
        self.shape_type = shape_type

        try:
            import shapefile
        except:
            raise Exception('ShapefileLayer requires pyshp')

        self.reader = shapefile.Reader(fname, encoding=encoding, encodingErrors=encodingErrors)
        self.worker = None
        self.queue = Queue.Queue()


    def invalidate(self, proj):
        self.painter = BatchPainter()
        self.hotspots = HotspotManager()
        self.painter.set_color(self.color)

        if self.worker:
            self.worker.stop()
            self.worker.join()
        self.queue = Queue.Queue()
        self.worker = ShapeLoadingThread(self.queue, self.reader, self.shape_type, proj)
        self.worker.start()


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        picked = self.hotspots.pick(mouse_x, mouse_y)
        if picked:
            ui_manager.tooltip(picked)

        while True:
            try:
                x, y, record = self.queue.get_nowait()
                self.painter.linestrip(x, y, self.linewidth, closed=True)
                if self.f_tooltip:
                    attr = {t[0][0]: t[1] for t in zip(self.reader.fields[1:], record)}
                    value = self.f_tooltip(attr)
                    if self.shape_type == 'bbox':
                        self.hotspots.add_rect(x.min(), y.min(), x.max()-x.min(), y.max()-y.min(), value)
                    else:
                        self.hotspots.add_poly(x, y, value)
            except Queue.Empty:
                break


class ShapeLoadingThread(Thread):

    def __init__(self, queue, reader, shape_type, proj):
        Thread.__init__(self)

        self.queue = queue
        self.reader = reader
        self.shape_type = shape_type
        self.proj = proj
        self.stop_flag = threading.Event()

        self.counter = 0

        self.daemon = True


    def stop(self):
        self.stop_flag.set()


    def run(self):
        while (self.counter < self.reader.numRecords) and (not self.stop_flag.is_set()):
            r = self.reader.shapeRecord(self.counter)
            if self.shape_type == 'bbox':
                top, left, bottom, right =  r.shape.bbox
                vertices = np.array([top, left, top, right, bottom, right, bottom, left]).reshape(-1,2)
            else:
                vertices = np.array(r.shape.points)
            x, y = self.proj.lonlat_to_screen(vertices[:,0], vertices[:,1])
            self.queue.put((x, y, r.record))
            self.counter += 1


class DelaunayLayer(BaseLayer):

    def __init__(self, data, line_color=None, line_width=2, cmap=None, max_lenght=100):
        """
        Draw a delaunay triangulation of the points

        :param data: data access object
        :param line_color: line color
        :param line_width: line width
        :param cmap: color map
        :param max_lenght: scaling constant for coloring the edges
        """
        self.data = data

        if cmap is None and line_color is None:
            raise Exception('need either cmap or line_color')

        if cmap is not None:
            cmap = colors.ColorMap(cmap, alpha=196)

        self.cmap = cmap
        self.line_color = line_color
        self.line_width = line_width
        self.max_lenght = max_lenght


    @staticmethod
    def _get_area(p):
        x1, y1, x2, y2, x3, y3 = p
        return 0.5*(x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2))



    def invalidate(self, proj):
        try:
            from scipy.spatial.qhull import Delaunay
        except ImportError:
            print('DelaunayLayer needs scipy >= 0.12')
            raise

        self.painter = BatchPainter()
        x, y = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])
        points = list(set(zip(x,y)))
        dela = Delaunay(points)

        edges = set()
        for tria in dela.vertices:
            edges.add((tria[0], tria[1]))
            edges.add((tria[1], tria[2]))
            edges.add((tria[2], tria[0]))

        allx0 = []
        ally0 = []
        allx1 = []
        ally1 = []
        colors = []

        for a, b in edges:
            x0, y0 = dela.points[a]
            x1, y1 = dela.points[b]

            allx0.append(x0)
            ally0.append(y0)
            allx1.append(x1)
            ally1.append(y1)

            if self.line_color:
                colors.append(self.line_color)
                colors.append(self.line_color)
            elif self.cmap:
                l = math.sqrt((x0 - x1)**2+(y0 - y1)**2)
                c = self.cmap.to_color(l, self.max_lenght, 'log')
                colors.append(c)
                colors.append(c)

        self.painter.lines(allx0, ally0, allx1, ally1, colors, width=self.line_width)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()


    def bbox(self):
        return BoundingBox.from_points(lons=self.data['lon'], lats=self.data['lat'])


class VoronoiLayer(BaseLayer):

    def __init__(self, data, line_color=None, line_width=2, f_tooltip=None, cmap=None, max_area=1e4, alpha=220):
        """
        Draw the voronoi tesselation of the points from data

        :param data: data access object
        :param line_color: line color
        :param line_width: line width
        :param f_tooltip: function to generate a tooltip on mouseover
        :param cmap: color map
        :param max_area: scaling constant to determine the color of the voronoi areas
        :param alpha: color alpha
        :return:
        """
        self.data = data

        if cmap is None and line_color is None:
            raise Exception('need either cmap or line_color')

        if cmap is not None:
            cmap = colors.ColorMap(cmap, alpha=alpha, levels=10)

        self.cmap = cmap
        self.line_color = line_color
        self.line_width = line_width
        self.f_tooltip = f_tooltip
        self.max_area = max_area


    # source: https://gist.github.com/pv/8036995
    @staticmethod
    def __voronoi_finite_polygons_2d(vor, radius=None):
        """
        Reconstruct infinite voronoi regions in a 2D diagram to finite
        regions.

        Parameters
        ----------
        vor : Voronoi
            Input diagram
        radius : float, optional
            Distance to 'points at infinity'.

        Returns
        -------
        regions : list of tuples
            Indices of vertices in each revised Voronoi regions.
        vertices : list of tuples
            Coordinates for revised Voronoi vertices. Same as coordinates
            of input vertices, with 'points at infinity' appended to the
            end.

        """

        if vor.points.shape[1] != 2:
            raise ValueError("Requires 2D input")

        new_regions = []
        new_vertices = vor.vertices.tolist()

        center = vor.points.mean(axis=0)
        if radius is None:
            radius = vor.points.ptp().max()

        # Construct a map containing all ridges for a given point
        all_ridges = {}
        for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
            all_ridges.setdefault(p1, []).append((p2, v1, v2))
            all_ridges.setdefault(p2, []).append((p1, v1, v2))

        # Reconstruct infinite regions
        for p1, region in enumerate(vor.point_region):
            vertices = vor.regions[region]

            if all(v >= 0 for v in vertices):
                # finite region
                new_regions.append(vertices)
                continue

            # reconstruct a non-finite region
            if p1 not in all_ridges:
                continue

            ridges = all_ridges[p1]
            new_region = [v for v in vertices if v >= 0]

            for p2, v1, v2 in ridges:
                if v2 < 0:
                    v1, v2 = v2, v1
                if v1 >= 0:
                    # finite ridge: already in the region
                    continue

                # Compute the missing endpoint of an infinite ridge

                t = vor.points[p2] - vor.points[p1] # tangent
                t /= np.linalg.norm(t)
                n = np.array([-t[1], t[0]])  # normal

                midpoint = vor.points[[p1, p2]].mean(axis=0)
                direction = np.sign(np.dot(midpoint - center, n)) * n
                far_point = vor.vertices[v2] + direction * radius

                new_region.append(len(new_vertices))
                new_vertices.append(far_point.tolist())

            # sort region counterclockwise
            vs = np.asarray([new_vertices[v] for v in new_region])
            c = vs.mean(axis=0)
            angles = np.arctan2(vs[:,1] - c[1], vs[:,0] - c[0])
            new_region = np.array(new_region)[np.argsort(angles)]

            # finish
            new_regions.append(new_region.tolist())

        return new_regions, np.asarray(new_vertices)


    # Area of a polygon: http://www.mathopenref.com/coordpolygonarea.html
    @staticmethod
    def _get_area(p):
        return 0.5 * abs(sum(x0*y1 - x1*y0
                             for ((x0, y0), (x1, y1)) in zip(p, p[1:] + [p[0]])))


    def invalidate(self, proj):
        try:
            from scipy.spatial.qhull import Voronoi
        except ImportError:
            print('VoronoiLayer needs scipy >= 0.12')
            raise

        x, y = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])
        points = zip(x,y)
        vor = Voronoi(points)

        regions, vertices = VoronoiLayer.__voronoi_finite_polygons_2d(vor)
        
        self.hotspots = HotspotManager()
        self.painter = BatchPainter()

        for idx, region in enumerate(regions):
            polygon = vertices[region]
            
            if self.line_color:
                self.painter.set_color(self.line_color)
                self.painter.linestrip(polygon[:,0], polygon[:,1], width=self.line_width, closed=True)
            if self.cmap:
                area = VoronoiLayer._get_area(polygon.tolist())
                area = max(area, 1)
                self.painter.set_color(self.cmap.to_color(area, self.max_area, 'log'))
                self.painter.poly(polygon[:,0], polygon[:,1])

            if self.f_tooltip:
                record = {k: self.data[k][idx] for k in self.data.keys()}
                self.hotspots.add_poly(polygon[:,0], polygon[:,1], self.f_tooltip(record))


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        picked = self.hotspots.pick(mouse_x, mouse_y)
        if picked:
            ui_manager.tooltip(picked)


    def bbox(self):
        return BoundingBox.from_points(lons=self.data['lon'], lats=self.data['lat'])


class MarkersLayer(BaseLayer):

    def __init__(self, data, marker, f_tooltip=None, marker_preferred_size=32):
        """
        Draw markers

        :param data: data access object
        :param marker: full filename of the marker image
        :param f_tooltip: function to generate a tooltip on mouseover
        :param marker_preferred_size: size in pixel for the marker images
        """
        self.data = data
        self.f_tooltip = f_tooltip
        self.marker_preferred_size = float(marker_preferred_size)
        self.marker = pyglet.image.load(marker)
        self.marker.anchor_x = self.marker.width / 2
        self.marker.anchor_y = self.marker.height / 2
        self.scale = self.marker_preferred_size / max(self.marker.width, self.marker.height)

        self.hotspots = HotspotManager()


    def invalidate(self, proj):
        self.painter = BatchPainter()
        x, y = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])

        if self.f_tooltip:
            for i in range(0, len(x)):
                record = {k: self.data[k][i] for k in self.data.keys()}
                self.hotspots.add_rect(x[i] - self.marker_preferred_size/2,
                                       y[i] - self.marker_preferred_size/2,
                                       self.marker_preferred_size,
                                       self.marker_preferred_size,
                                       self.f_tooltip(record))

        self.painter.sprites(self.marker, x, y, self.scale)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()

        picked = self.hotspots.pick(mouse_x, mouse_y)
        if picked:
            ui_manager.tooltip(picked)


    def bbox(self):
        return BoundingBox.from_points(lons=self.data['lon'], lats=self.data['lat'])


class KDELayer(BaseLayer):

    def __init__(self, values, bw, cmap='hot', method='hist', scaling='sqrt', alpha=220,
                 cut_below=None, clip_above=None, binsize=1, cmap_levels=10, show_colorbar=False):
        """
        Kernel density estimation visualization

        :param data: data access object
        :param bw: kernel bandwidth (in screen coordinates)
        :param cmap: colormap
        :param method: if kde use KDEMultivariate from statsmodel, which provides a more accurate but much slower estimation.
            If hist, estimates density applying gaussian smoothing on a 2D histogram, which is much faster but less accurate
        :param scaling: colorscale, lin log or sqrt
        :param alpha: color alpha
        :param cut_below: densities below cut_below are not drawn
        :param clip_above: defines the max value for the colorscale
        :param binsize: size of the bins for hist estimator
        :param cmap_levels: discretize colors into cmap_levels
        :param show_colorbar: show colorbar
        """
        self.values = values
        self.bw = bw
        self.cmap = colors.ColorMap(cmap, alpha=alpha, levels=cmap_levels)
        self.method = method
        self.scaling = scaling
        self.cut_below = cut_below
        self.clip_above = clip_above
        self.binsize = binsize
        self.show_colorbar = show_colorbar


    def _get_grid(self, proj):
        west, north = proj.lonlat_to_screen([proj.bbox().west], [proj.bbox().north])
        east, south = proj.lonlat_to_screen([proj.bbox().east], [proj.bbox().south])
        xgrid = np.arange(west, east, self.binsize)
        ygrid = np.arange(south, north, self.binsize)
        return xgrid, ygrid


    def invalidate(self, proj):
        self.painter = BatchPainter()
        xv, yv = proj.lonlat_to_screen(self.values['lon'], self.values['lat'])

        rects_vertices = []
        rects_colors = []

        if self.method == 'kde':
            try:
                import statsmodels.api as sm
            except:
                raise Exception('KDE requires statsmodel')

            kde_res = sm.nonparametric.KDEMultivariate(data=[xv, yv], var_type='cc', bw=self.bw)
            xgrid, ygrid = self._get_grid(proj)
            xmesh, ymesh = np.meshgrid(xgrid,ygrid)
            grid_coords = np.append(xmesh.reshape(-1,1), ymesh.reshape(-1,1),axis=1)
            z = kde_res.pdf(grid_coords.T)
            z = z.reshape(len(ygrid), len(xgrid))

            # np.save('z.npy', z)
            # z = np.load('z.npy')

            print('smallest non-zero density:', z[z > 0][0])
            print('max density:', z.max())

            if self.cut_below is None:
                zmin = z[z > 0][0]
            else:
                zmin = self.cut_below

            if self.clip_above is None:
                zmax = z.max()
            else:
                zmax = self.clip_above

            for ix in range(len(xgrid)-1):
                for iy in range(len(ygrid)-1):
                    if z[iy, ix] > zmin:
                        rects_vertices.append((xgrid[ix], ygrid[iy], xgrid[ix+1], ygrid[iy+1]))
                        rects_colors.append(self.cmap.to_color(z[iy, ix], zmax, self.scaling))
        elif self.method == 'hist':
            try:
                from scipy.ndimage import gaussian_filter
            except:
                raise Exception('KDE requires scipy')

            xgrid, ygrid = self._get_grid(proj)
            H, _, _ = np.histogram2d(yv, xv, bins=(ygrid, xgrid))
            
            if H.sum() == 0:
                print('no data in current view')
                return

            H = gaussian_filter(H, sigma=self.bw)
            print('smallest non-zero count', H[H > 0][0])
            print('max count:', H.max())

            if self.cut_below is None:
                Hmin = H[H > 0][0]
            else:
                Hmin = self.cut_below

            if self.clip_above is None:
                self.Hmax = H.max()
            else:
                self.Hmax = self.clip_above

            if self.scaling == 'ranking':
                from statsmodels.distributions.empirical_distribution import ECDF
                ecdf = ECDF(H.flatten())

            for ix in range(len(xgrid)-2):
                for iy in range(len(ygrid)-2):
                    if H[iy, ix] > Hmin:
                        rects_vertices.append((xgrid[ix], ygrid[iy], xgrid[ix+1], ygrid[iy+1]))
                        if self.scaling == 'ranking':
                            rects_colors.append(self.cmap.to_color(ecdf(H[iy, ix]) - ecdf(Hmin), 1 - ecdf(Hmin), 'lin'))
                        else:
                            rects_colors.append(self.cmap.to_color(H[iy, ix], self.Hmax, self.scaling))
        else:
            raise Exception('method not supported')

        self.painter.batch_rects(rects_vertices, rects_colors)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        if self.show_colorbar:
            ui_manager.add_colorbar(self.cmap, self.Hmax, self.scaling)


class ConvexHullLayer(BaseLayer):

    def __init__(self, data, col, fill=True, point_size=4):
        """
        Convex hull for a set of points

        :param data: points
        :param col: color
        :param fill: whether to fill the convexhull polygon or not
        :param point_size: size of the points on the convexhull. Points are not rendered if None
        """
        self.data = data
        self.col = col
        self.fill = fill
        self.point_size=point_size


    def invalidate(self, proj):
        self.painter = BatchPainter()
        self.painter.set_color(self.col)
        x, y = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])
        if len(x) >= 3:
            self.painter.convexhull(x, y, self.fill)
        else:
            self.painter.linestrip(x, y)

        if self.point_size > 0:
            self.painter.points(x, y, self.point_size)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()


class GridLayer(BaseLayer):

    def __init__(self, lon_edges, lat_edges, values, cmap, alpha=255, vmin=None, vmax=None, levels=10, 
            colormap_scale='lin', show_colorbar=True):
        """
        Values over a uniform grid
        
        :param lon_edges: longitude edges
        :param lat_edges: latitude edges
        :param values: matrix representing values on the grid 
        :param cmap: colormap name
        :param alpha: color alpha
        :param vmin: minimum value for the colormap
        :param vmax: maximum value for the colormap
        :param levels: number of levels for the colormap
        :param colormap_scale: colormap scale
        :param show_colorbar: show the colorbar in the UI
        """
        self.lon_edges = lon_edges
        self.lat_edges = lat_edges
        self.values = values
        self.cmap = colors.ColorMap(cmap, alpha=alpha, levels=levels)
        self.colormap_scale = colormap_scale
        self.show_colorbar = show_colorbar

        if vmin:
            self.vmin = vmin
        else:
            self.vmin = 0

        if vmax:
            self.vmax = vmax
        else:
            self.vmax = self.values[~np.isnan(self.values)].max()


    def invalidate(self, proj):
        self.painter = BatchPainter()
        xv, yv = proj.lonlat_to_screen(self.lon_edges, self.lat_edges)

        rects = []
        cols = []
        for ix in range(len(xv)-1):
            for iy in range(len(yv)-1):
                d = self.values[iy, ix]
                if d > self.vmin:
                    rects.append((xv[ix], yv[iy], xv[ix+1], yv[iy+1]))
                    cols.append(self.cmap.to_color(d, self.vmax, self.colormap_scale))

        self.painter.batch_rects(rects, cols)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        if self.show_colorbar:
            ui_manager.add_colorbar(self.cmap, self.vmax, self.colormap_scale)


    def bbox(self):
        return BoundingBox(north=self.lat_edges[-1], south=self.lat_edges[0], 
                           west=self.lon_edges[0], east=self.lon_edges[-1])


class GeoJSONLayer(BaseLayer):

    def __init__(self, geojson_or_fname, color='b', linewidth=1, fill=False, f_tooltip=None):
        self.color = color
        self.linewidth = linewidth
        self.fill = fill
        self.f_tooltip = f_tooltip

        if type(geojson_or_fname) == str:
            with open(geojson_or_fname) as fin:
                self.data = json.load(fin)
        elif type(geojson_or_fname) == dict:
            self.data = geojson_or_fname
        else:
            raise Exception('must provide either dict or filename')

        self.boundingbox = None
        
        for feature in self.data['features']:
            if feature['geometry'] is None:
                print('feature without geometry data: %s' % feature['properties']['NAME'])
                continue
                
            if feature['geometry']['type'] == 'Polygon':
                for poly in feature['geometry']['coordinates']: 
                    poly = np.array(poly)
                    self.__update_bbox(poly[:,0], poly[:,1])
            elif feature['geometry']['type'] == 'MultiPolygon':
                for multipoly in feature['geometry']['coordinates']:
                    for poly in multipoly: 
                        poly = np.array(poly)
                        self.__update_bbox(poly[:,0], poly[:,1])
            elif feature['geometry']['type'] == 'Point':
                lon,lat = feature['geometry']['coordinates']
                self.__update_bbox(np.array([lon]), np.array([lat]))
            elif feature['geometry']['type'] == 'LineString':
                line = np.array(feature['geometry']['coordinates'])
                self.__update_bbox(line[:,0], line[:,1])


    def __update_bbox(self, lon, lat):
        if self.boundingbox is None:
            self.boundingbox = BoundingBox(north=lat.max(), south=lat.min(), west=lon.min(), east=lon.max())
        else:
            self.boundingbox = BoundingBox(
                                    north=max(self.boundingbox.north, lat.max()),
                                    south=min(self.boundingbox.south, lat.min()),
                                    west=min(self.boundingbox.west, lon.min()),
                                    east=max(self.boundingbox.east, lon.max()))


    def invalidate(self, proj):
        self.painter = BatchPainter()
        self.hotspots = HotspotManager()

        for feature in self.data['features']:
            if isfunction(self.color):
                self.painter.set_color(self.color(feature['properties']))
            else:
                self.painter.set_color(self.color)

            if feature['geometry']['type'] == 'Polygon':
                for poly in feature['geometry']['coordinates']: 
                    poly = np.array(poly)
                    x, y = proj.lonlat_to_screen(poly[:,0], poly[:,1])
                    if self.fill:
                        self.painter.poly(x, y)
                    else:
                        self.painter.linestrip(x, y, self.linewidth, closed=True)

                    if self.f_tooltip:
                        self.hotspots.add_poly(x, y, self.f_tooltip(feature['properties']))

            elif feature['geometry']['type'] == 'MultiPolygon':
                for multipoly in feature['geometry']['coordinates']:
                    for poly in multipoly: 
                        poly = np.array(poly)
                        x, y = proj.lonlat_to_screen(poly[:,0], poly[:,1])
                        if self.fill:
                            self.painter.poly(x, y)
                        else:
                            self.painter.linestrip(x, y, self.linewidth, closed=True)

                        if self.f_tooltip:
                            self.hotspots.add_poly(x, y, self.f_tooltip(feature['properties']))
            
            elif feature['geometry']['type'] == 'Point':
                lon,lat = feature['geometry']['coordinates']
                x, y = proj.lonlat_to_screen(np.array([lon]), np.array([lat]))
                self.painter.points(x, y)
            elif feature['geometry']['type'] == 'LineString':
                line = np.array(feature['geometry']['coordinates'])
                x, y = proj.lonlat_to_screen(line[:,0], line[:,1])
                self.painter.linestrip(x, y, self.linewidth, closed=False)
            else:
                print('unknow geometry %s' % feature['geometry']['type'])


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        picked = self.hotspots.pick(mouse_x, mouse_y)
        if picked:
            ui_manager.tooltip(picked)


    def bbox(self):
        if self.boundingbox:
            return self.boundingbox
        else:
            return BoundingBox.WORLD


class LabelsLayer(BaseLayer):


    def __init__(self, data, label_column, color=None, font_name=FONT_NAME, font_size=14, anchor_x='left', anchor_y='top'):
        """Create a layer with a text label for each sample

        :param data: data access object
        :param label_column: column in the data access object where the labels text is stored
        :param color: color
        :param font_name: font name
        :param font_size: font size
        :param anchor_x: anchor x
        :param anchor_y: anchor y
        """
        self.data = data
        self.label_column = label_column
        self.color = color
        self.font_name = font_name
        self.font_size = font_size
        self.anchor_x = anchor_x
        self.anchor_y = anchor_y
        if self.color is None:
            self.color = [255,0,0]


    def invalidate(self, proj):
        self.painter = BatchPainter()
        x, y = proj.lonlat_to_screen(self.data['lon'], self.data['lat'])
        self.painter.set_color(self.color)
        self.painter.labels(x, y, self.data[self.label_column], 
                                    font_name=self.font_name,
                                    font_size=self.font_size,
                                    anchor_x=self.anchor_x, 
                                    anchor_y=self.anchor_y)


    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter.batch_draw()
        

    def bbox(self):
        return BoundingBox.from_points(lons=self.data['lon'], lats=self.data['lat'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from Queue import Queue
from threading import Thread
from pyglet.window import mouse
import time
from pyglet.gl import *
import math
import numpy as np
import os
import random
import urllib2
import pyglet
from os.path import expanduser
from geoplotlib import utils
from geoplotlib.utils import BoundingBox, parse_raw_str


VERT_PER_POINT = 2
FPS = 30
TILE_SIZE = 256
MIN_ZOOM = 2
MAX_ZOOM = 20
KEYBOARD_PAN = 0.2
TOTAL_INVALIDATE_DELAY = 50
FONT_COLOR = (0,0,0,255)
FONT_NAME = 'Helvetica'
FONT_SCALING = 1./100


class UiManager:

    class Colorbar():

        def __init__(self, cmap, vmax, colormap_scale, font_size, size=.5):
            self.cmap = cmap
            self.vmax = vmax
            self.colormap_scale = colormap_scale
            self.font_size = font_size
            self.size = size


        def draw(self, painter):
            total_h = SCREEN_H*self.size
            step = total_h / self.cmap.levels
            bar_w = SCREEN_W/25

            lab = pyglet.text.Label('',
                               color=FONT_COLOR,
                               font_name=FONT_NAME,
                               font_size=int(.8*self.font_size),
                               x=SCREEN_W, y=SCREEN_H,
                               anchor_x='right', anchor_y='center')

            edges, colors = self.cmap.get_boundaries(self.vmax, self.colormap_scale)

            for i in range(self.cmap.levels+1):
                if i < self.cmap.levels:
                    painter.set_color(colors[i][:-1])
                    painter.rect(SCREEN_W-2*bar_w/2, SCREEN_H-total_h*1.5+step*i, 
                                 SCREEN_W-bar_w/2, SCREEN_H-total_h*1.5+step*(i+1))
                lab.x = SCREEN_W-2*bar_w/2*1.1
                lab.y = SCREEN_H-total_h*1.5+step*i
                if self.colormap_scale == 'log':
                    lab.text = '%.2E' % edges[i]
                else:
                    lab.text = '%d' % edges[i]
                lab.draw()


    def __init__(self):
        self.font_size = int(SCREEN_W*FONT_SCALING)
        self.padding = 2
        self.labels = {}

        self.labels['status'] = pyglet.text.Label('',
                                       color=FONT_COLOR,
                                       font_name=FONT_NAME,
                                       font_size=self.font_size,
                                       x=20, y=10,
                                       anchor_x='left', anchor_y='bottom')

        self.labels['tooltip'] = pyglet.text.Label('',
                                       color=FONT_COLOR,
                                       font_name=FONT_NAME,
                                       font_size=self.font_size,
                                       x=SCREEN_W, y=SCREEN_H,
                                       anchor_x='left', anchor_y='bottom')

        self.labels['info'] = pyglet.text.Label('',
                                       color=FONT_COLOR,
                                       font_name=FONT_NAME,
                                       font_size=self.font_size,
                                       x=SCREEN_W, y=SCREEN_H,
                                       anchor_x='right', anchor_y='top')

        self.colorbar = None


    def tooltip(self, text):
        self.labels['tooltip'].text = parse_raw_str(text)


    def status(self, text):
        self.labels['status'].text = parse_raw_str(text)


    def info(self, text):
        self.labels['info'].text = parse_raw_str(text)


    @staticmethod
    def get_label_bbox(label):

        if label.anchor_x == 'left':
            left = label.x
        elif label.anchor_x == 'right':
            left = label.x - label.content_width

        if label.anchor_y == 'bottom':
            top = label.y
        elif label.anchor_y == 'top':
            top = label.y - label.content_height

        return left, top, left + label.content_width, top + label.content_height


    def draw_label_background(self, label, painter):
        if len(label.text) > 0:
            left, top, right, bottom = UiManager.get_label_bbox(label)
            painter.rect(left - self.padding, top - self.padding, right + self.padding, bottom + self.padding)


    def draw(self, mouse_x, mouse_y):
        painter = BatchPainter()

        if self.colorbar:
            self.colorbar.draw(painter)

        painter.set_color([255,255,255])
        self.labels['tooltip'].x = mouse_x
        self.labels['tooltip'].y = mouse_y
        for l in self.labels.values():
            self.draw_label_background(l, painter)
        painter.batch_draw()
        for l in self.labels.values():
            l.draw()


    def clear(self):
        for l in self.labels.values():
            l.text = ''


    def add_colorbar(self, cmap, vmax, colormap_scale):
        self.colorbar = UiManager.Colorbar(cmap, vmax, colormap_scale, self.font_size)


class GeoplotlibApp(pyglet.window.Window):

    def __init__(self, geoplotlib_config):
        super(GeoplotlibApp, self).__init__(geoplotlib_config.screen_w, geoplotlib_config.screen_h,
                                            fullscreen=False, caption='geoplotlib')
        global SCREEN_W, SCREEN_H
        SCREEN_W = geoplotlib_config.screen_w
        SCREEN_H = geoplotlib_config.screen_h

        self.geoplotlib_config = geoplotlib_config

        self.ticks = 0
        self.ui_manager = UiManager()
        self.proj = Projector()
        self.map_layer = MapLayer(geoplotlib_config.tiles_provider, skipdl=False)

        self.scroll_delay = 0
        self.invalidate_delay = 0
        self.drag_x = self.drag_y = 0
        self.dragging = False
        self.drag_start_timestamp = 0
        self.mouse_x = self.mouse_y = 0
        self.show_map = True
        self.show_layers = True
        self.show_coordinates = False

        glEnable(GL_LINE_SMOOTH)
        glEnable(GL_POLYGON_SMOOTH)
        # glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
        # glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        pyglet.clock.schedule_interval(self.on_update, 1. / FPS)


    def on_draw(self):
        self.clear()

        # needed to avoid diagonal artifacts on the tiles
        glDisable(GL_LINE_SMOOTH)
        glDisable(GL_POLYGON_SMOOTH)

        self.ui_manager.clear()

        if self.show_map:
            self.map_layer.draw(self.proj)

            if self.geoplotlib_config.map_alpha < 255:
                painter = BatchPainter()
                painter.set_color([0,0,0, 255 - self.geoplotlib_config.map_alpha])
                painter.rect(0,0,SCREEN_W, SCREEN_H)
                painter.batch_draw()


        if abs(self.drag_x) > 1e-3 or abs(self.drag_y) > 1e-3:
            self.drag_x *= 0.93
            self.drag_y *= 0.93

            if self.dragging == False:
                self.proj.pan(self.drag_x, self.drag_y)

        if self.scroll_delay > 0:
            self.scroll_delay -= 1

        if self.invalidate_delay > 0:
            self.invalidate_delay -= 1
        if self.invalidate_delay == 1:
            for l in self.geoplotlib_config.layers:
                l.invalidate(self.proj)
        if self.show_layers and self.invalidate_delay == 0:
            if self.geoplotlib_config.smoothing:
                glEnable(GL_LINE_SMOOTH)
                glEnable(GL_POLYGON_SMOOTH)

            glPushMatrix()
            glTranslatef(-self.proj.xtile * TILE_SIZE, self.proj.ytile * TILE_SIZE, 0)
            for l in self.geoplotlib_config.layers:
                l.draw(self.proj,
                       self.mouse_x + self.proj.xtile * TILE_SIZE,
                       SCREEN_H - self.mouse_y - self.proj.ytile * TILE_SIZE,
                       self.ui_manager)
            glPopMatrix()

            #self.ui_manager.status('T: %.1f, FPS:%d' % (self.ticks / 1000., pyglet.clock.get_fps()))
            if self.show_coordinates:
                self.ui_manager.status('%.6f %.6f' % self.proj.screen_to_latlon(self.mouse_x, SCREEN_H - self.mouse_y))

        if self.invalidate_delay == 2:
            self.ui_manager.status('rendering...')

        attribution = pyglet.text.Label(self.map_layer.attribution,
                                       color=FONT_COLOR,
                                       font_name=FONT_NAME,
                                       font_size=int(.8*SCREEN_W*FONT_SCALING),
                                       x=SCREEN_W-int(.2*SCREEN_W*FONT_SCALING), 
                                       y=int(1.2*SCREEN_W*FONT_SCALING),
                                       anchor_x='right', anchor_y='top')
        attribution.draw()
        self.ui_manager.draw(self.mouse_x, SCREEN_H - self.mouse_y)

        if self.geoplotlib_config.savefig is not None:
            GeoplotlibApp.screenshot(self.geoplotlib_config.savefig + '.png')
            pyglet.app.exit()


    def on_mouse_motion(self, x, y, dx, dy):
        self.mouse_x = x
        self.mouse_y = SCREEN_H - y


    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        if buttons & mouse.LEFT:
            self.drag_start_timestamp = self.ticks
            self.drag_x = -1. * dx / TILE_SIZE
            self.drag_y = -1. * dy / TILE_SIZE
            self.proj.pan(self.drag_x, self.drag_y)
            if self.invalidate_delay > 0:
                    self.invalidate_delay = TOTAL_INVALIDATE_DELAY


    def on_mouse_release(self, x, y, buttons, modifiers):
        if buttons == mouse.LEFT:
            self.dragging = False
            if self.ticks - self.drag_start_timestamp > 200:
                self.drag_x = self.drag_y = 0


    def on_mouse_press(self, x, y, buttons, modifiers):
        if buttons == mouse.LEFT:
            if not self.dragging:
                self.dragging = True
                self.drag_start_timestamp = self.ticks
                self.drag_x = self.drag_y = 0
                if self.invalidate_delay > 0:
                    self.invalidate_delay = TOTAL_INVALIDATE_DELAY


    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        if self.scroll_delay == 0:
            if scroll_y < 0:
                self.proj.zoomin(self.mouse_x, self.mouse_y)
                self.invalidate_delay = TOTAL_INVALIDATE_DELAY
                self.scroll_delay = 3
            elif scroll_y > 0:
                self.proj.zoomout(self.mouse_x, self.mouse_y)
                self.invalidate_delay = TOTAL_INVALIDATE_DELAY
                self.scroll_delay = 3


    def on_key_release(self, symbol, modifiers):
        if symbol == pyglet.window.key.P:
            fname = '%d.png' % (time.time()*1000)
            GeoplotlibApp.screenshot(fname)
            print(fname + ' saved')
        elif symbol == pyglet.window.key.M:
            self.show_map = not self.show_map
        elif symbol == pyglet.window.key.L:
            self.show_layers = not self.show_layers
        elif symbol == pyglet.window.key.I:
            self.proj.zoomin(SCREEN_W/2, SCREEN_H/2)
            self.invalidate_delay = TOTAL_INVALIDATE_DELAY
        elif symbol == pyglet.window.key.O:
            self.proj.zoomout(SCREEN_W/2, SCREEN_H/2)
            self.invalidate_delay = TOTAL_INVALIDATE_DELAY
        elif symbol == pyglet.window.key.R:
            # hack to force invalidate
            self.invalidate_delay = 3
        elif symbol == pyglet.window.key.A:
            self.proj.pan(-KEYBOARD_PAN, 0)
        elif symbol == pyglet.window.key.D:
            self.proj.pan(+KEYBOARD_PAN, 0)
        elif symbol == pyglet.window.key.W:
            self.proj.pan(0, +KEYBOARD_PAN)
        elif symbol == pyglet.window.key.S:
            self.proj.pan(0, -KEYBOARD_PAN)
        elif symbol == pyglet.window.key.B:
            print(self.proj.bbox())
        elif symbol == pyglet.window.key.C:
            self.show_coordinates = not self.show_coordinates
        else:
            for l in self.geoplotlib_config.layers:
                need_invalidate = l.on_key_release(symbol, modifiers)
                if need_invalidate:
                    l.invalidate(self.proj)


    @staticmethod
    def screenshot(fname):
        glPixelTransferf(gl.GL_ALPHA_BIAS, 1.0)
        image = pyglet.image.ColorBufferImage(0, 0, SCREEN_W, SCREEN_H)
        image.save(fname)
        glPixelTransferf(gl.GL_ALPHA_BIAS, 0.0)


    def on_update(self, dt):
        self.ticks += dt*1000


    def start(self):
        #pyglet.options['debug_gl'] = False
        if self.geoplotlib_config.bbox is not None:
            self.proj.fit(self.geoplotlib_config.bbox, force_zoom=self.geoplotlib_config.requested_zoom)
        elif len(self.geoplotlib_config.layers) > 0:
            self.proj.fit(BoundingBox.from_bboxes([l.bbox() for l in self.geoplotlib_config.layers]),
                          force_zoom=self.geoplotlib_config.requested_zoom)
        for l in self.geoplotlib_config.layers:
            l.invalidate(self.proj)

        pyglet.app.run()


def _flatten_xy(x, y):
        return np.vstack((x, y)).T.flatten()


class BatchPainter:
    """
    This class batches OpenGL calls. The usage pattern is to instantiate a BatchPainter,
     perform all the drawing and finally render using batch_draw
    """

    def __init__(self):
        self._batch = pyglet.graphics.Batch()
        self._color = [0, 0, 255, 255]
        self._sprites = []
        self._labels = []


    def set_color(self, color):
        if color == 'k' or color == 'black':
            self._color = [0,0,0,255]
        elif color == 'w' or color == 'white':
            self._color = [255,255,255,255]
        elif color == 'r' or color == 'red':
            self._color = [255,0,0,255]
        elif color == 'g' or color == 'green':
            self._color = [0,255,0,255]
        elif color == 'b' or color == 'blue':
            self._color = [0,0,255,255]
        elif len(color) == 4:
            for c in color:
                if c < 0 or c > 255:
                    raise Exception('color components must be between 0 and 255')
            self._color = color
        elif len(color) == 3:
            for c in color:
                if c < 0 or c > 255:
                    raise Exception('color components must be between 0 and 255')
            self._color = color + [255]
        else:
            raise Exception('invalid color format')


    def lines(self, x0, y0, x1, y1, colors=None, width=1.0):
        glLineWidth(width)
        x = _flatten_xy(x0, x1)
        y = _flatten_xy(y0, y1)
        vertices = _flatten_xy(x, y)
        if colors is None:
            colors = self._color * int(len(vertices)/VERT_PER_POINT)

        self._batch.add(int(len(vertices)/VERT_PER_POINT), GL_LINES, None,
                      ('v2f', vertices),
                      ('c4B', np.array(colors).flatten()))


    def linestrip(self, x, y, width=1.0, closed=False):
        glLineWidth(width)
        vertices = _flatten_xy(x, y)
        indices = [i // 2 for i in range(len(vertices))]
        indices = indices[1:-1]
        if closed:
            indices.append(indices[-1])
            indices.append(indices[0])

        self._batch.add_indexed(int(len(vertices)/VERT_PER_POINT), GL_LINES, None,
                      indices,
                      ('v2f', vertices),
                      ('c4B', self._color * int(len(vertices)/VERT_PER_POINT)))


    def poly(self, x, y, width=1.0):
        glLineWidth(width)
        vertices = _flatten_xy(x, y)
        indices = []
        for i in range(1, len(x) - 1):
            indices.append(0)
            indices.append(i)
            indices.append(i+1)

        self._batch.add_indexed(int(len(vertices)/VERT_PER_POINT), GL_TRIANGLES, None,
                      indices,
                      ('v2f', vertices),
                      ('c4B', self._color * int(len(vertices)/VERT_PER_POINT)))


    def triangle(self, vertices):
        self._batch.add(int(len(vertices)/VERT_PER_POINT), GL_TRIANGLES, None,
                      ('v2f', vertices),
                      ('c4B', self._color * int(len(vertices)/VERT_PER_POINT)))


    def circle(self, cx, cy, r, width=2.0):
        glLineWidth(width)

        precision = int(10*math.log(r))

        vertices = []
        for alpha in np.linspace(0, 6.28, precision):
            vertices.append(cx + r * math.cos(alpha))
            vertices.append(cy + r * math.sin(alpha))

        indices = []
        for i in range(precision - 1):
            indices.append(i)
            indices.append(i+1)
        indices.append(precision-1)
        indices.append(0)

        self._batch.add_indexed(int(len(vertices)/VERT_PER_POINT), GL_LINES, None,
                      indices,
                      ('v2f', vertices),
                      ('c4B', self._color * int(len(vertices)/VERT_PER_POINT)))


    def circle_filled(self, cx, cy, r):
        vertices = []
        vertices.append(cx)
        vertices.append(cy)

        precision = int(10*math.log(r))

        for alpha in np.linspace(0, 6.28, precision):
            vertices.append(cx + r * math.cos(alpha))
            vertices.append(cy + r * math.sin(alpha))

        indices = []
        for i in range(1, precision):
            indices.append(0)
            indices.append(i)
            indices.append(i+1)
        indices.append(0)
        indices.append(precision)
        indices.append(1)

        self._batch.add_indexed(int(len(vertices)/VERT_PER_POINT), GL_TRIANGLES, None,
                      indices,
                      ('v2f', vertices),
                      ('c4B', self._color * int(len(vertices)/VERT_PER_POINT)))


    def points(self, x, y, point_size=10, rounded=False):
        glPointSize(point_size)
        if rounded:
            glEnable(GL_POINT_SMOOTH)
        else:
            glDisable(GL_POINT_SMOOTH)

        vertices = np.vstack((x, y)).T.flatten()

        self._batch.add(int(len(vertices)/VERT_PER_POINT), GL_POINTS, None,
                        ('v2f', vertices),
                        ('c4B', self._color * int(len(vertices)/VERT_PER_POINT)))


    def rect(self, left, top, right, bottom):
        self.triangle([left, top, right, top, right, bottom, right, bottom, left, top, left, bottom])


    def batch_rects(self, rects_vertices, rects_colors):
        triangles = []
        colors = []
        for i in range(len(rects_vertices)):
            r = rects_vertices[i]
            c = rects_colors[i]
            left, top, right, bottom = r
            triangles.extend([left, top, right, top, right, bottom, right, bottom, left, top, left, bottom])
            colors.extend(c * 6)

        self._batch.add(int(len(triangles)/VERT_PER_POINT), GL_TRIANGLES, None,
                      ('v2f', triangles),
                      ('c4B', colors))


    def sprites(self, image, x, y, scale=1.0):
        from pyglet.sprite import Sprite
        for i in range(len(x)):
            sprite = Sprite(image, batch=self._batch)
            sprite.x = x[i]
            sprite.y = y[i]
            sprite.scale = scale
            self._sprites.append(sprite)


    def labels(self, x, y, texts, font_name=FONT_NAME, font_size=14, anchor_x='left', anchor_y='top'):
        for i in range(len(x)):
            lab = pyglet.text.Label(parse_raw_str(texts if type(texts) == str else texts[i]),
                                    batch=self._batch,
                                    color=self._color,
                                    font_name=font_name,
                                    font_size=font_size,
                                    x=x[i], y=y[i],
                                    anchor_x=anchor_x, 
                                    anchor_y=anchor_y)
            self._labels.append(lab)


    def convexhull(self, x, y, fill=False, smooth=False):
        try:
            from scipy.spatial import ConvexHull
            from scipy.spatial.qhull import QhullError
        except:
            raise Exception('ConvexHull requires scipy')

        if len(x) < 3:
            raise Exception('convexhull requires at least 3 points')


        points = np.vstack((x,y)).T
        try:
            hull = ConvexHull(points)
            xhull = points[hull.vertices,0]
            yhull = points[hull.vertices,1]

            if smooth:
                xhull, yhull = self.__generate_spline(xhull, yhull, closed=True)

            if fill:
                self.poly(xhull,yhull)
            else:
                self.linestrip(xhull, yhull, 3, closed=True)

        except QhullError as qerr:
            self.linestrip(x, y, 3, closed=False)


    def __generate_spline(self, x, y, closed=False, steps=20):
        """
        catmullrom spline
        http://www.mvps.org/directx/articles/catmull/
        """

        if closed:
            x = x.tolist()
            x.insert(0, x[-1])
            x.append(x[1])
            x.append(x[2])

            y = y.tolist()
            y.insert(0, y[-1])
            y.append(y[1])
            y.append(y[2])

        points = np.vstack((x,y)).T

        curve = []

        if not closed:
            curve.append(points[0])

        for j in range(1, len(points)-2):
            for s in range(steps):
                t = 1. * s / steps
                p0, p1, p2, p3 = points[j-1], points[j], points[j+1], points[j+2]
                pnew = 0.5 *((2 * p1) + (-p0 + p2) * t + (2*p0 - 5*p1 + 4*p2 - p3) * t**2 + (-p0 + 3*p1- 3*p2 + p3) * t**3)
                curve.append(pnew)

        if not closed:
            curve.append(points[-1])

        curve = np.array(curve)
        return curve[:, 0], curve[:, 1]


    def spline(self, x, y, width=3):
        xcurve, ycurve = self.__generate_spline(x, y, closed=False)
        self.linestrip(xcurve, ycurve, width)


    def batch_draw(self):
        self._batch.draw()


class Projector():

    def __init__(self):
        self.tiles_horizontally = 1.*SCREEN_W / TILE_SIZE
        self.tiles_vertically = 1.*SCREEN_H / TILE_SIZE
        self.fit(BoundingBox.WORLD)


    def set_to(self, north, west, zoom):
        self.zoom = zoom
        self.xtile, self.ytile = self.deg2num(north, west, zoom)


    def fit(self, bbox, max_zoom=MAX_ZOOM, force_zoom=None):
        """
        Fits the projector to a BoundingBox
        :param bbox: BoundingBox
        :param max_zoom: max zoom allowed
        :param force_zoom: force this specific zoom value even if the whole bbox does not completely fit
        """

        BUFFER_FACTOR = 1.1
        
        if force_zoom is not None:
            self.zoom = force_zoom
        else:
            for zoom in range(max_zoom, MIN_ZOOM-1, -1):
                self.zoom = zoom
                left, top = self.lonlat_to_screen([bbox.west], [bbox.north])
                right, bottom = self.lonlat_to_screen([bbox.east], [bbox.south])
                if (top - bottom < SCREEN_H*BUFFER_FACTOR) and (right - left < SCREEN_W*BUFFER_FACTOR):
                    break

        west_tile, north_tile = self.deg2num(bbox.north, bbox.west, self.zoom)
        east_tile, south_tile = self.deg2num(bbox.south, bbox.east, self.zoom)
        self.xtile = west_tile - self.tiles_horizontally/2. + (east_tile - west_tile)/2
        self.ytile = north_tile - self.tiles_vertically/2. + (south_tile - north_tile)/2
        self.calculate_viewport_size()


    @staticmethod
    def deg2num(lat_deg, lon_deg, zoom):
        lat_rad = math.radians(lat_deg)
        n = 2.0 ** zoom
        xtile = (lon_deg + 180.0) / 360.0 * n
        ytile = (1.0 - math.log(math.tan(lat_rad) + (1 / math.cos(lat_rad))) / math.pi) / 2.0 * n
        return (xtile, ytile)


    @staticmethod
    def num2deg(xtile, ytile, zoom):
        n = 2.0 ** zoom
        lon_deg = xtile / n * 360.0 - 180.0
        lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
        lat_deg = math.degrees(lat_rad)
        return (lat_deg, lon_deg)


    def bbox(self):
        north, west = self.num2deg(self.xtile, self.ytile, self.zoom)
        south, east = self.num2deg(self.xtile + self.tiles_horizontally, self.ytile + self.tiles_vertically, self.zoom)
        return BoundingBox(north=north, west=west, south=south, east=east)


    def pan(self, deltax, deltay):
        self.xtile += deltax
        self.ytile -= deltay


    def zoomin(self, mouse_x, mouse_y):
        mouse_lat, mouse_lon = self.screen_to_latlon(mouse_x, mouse_y)
        self.zoom = min(self.zoom + 1, MAX_ZOOM)
        self.xtile, self.ytile = self.deg2num(mouse_lat, mouse_lon, self.zoom)
        self.xtile -= 1. * mouse_x / TILE_SIZE
        self.ytile -= 1. * mouse_y / TILE_SIZE
        self.calculate_viewport_size()


    def zoomout(self, mouse_x, mouse_y):
        mouse_lat, mouse_lon = self.screen_to_latlon(mouse_x, mouse_y)
        self.zoom = max(self.zoom - 1, MIN_ZOOM)
        self.xtile, self.ytile = self.deg2num(mouse_lat, mouse_lon, self.zoom)
        self.xtile -= 1. * mouse_x / TILE_SIZE
        self.ytile -= 1. * mouse_y / TILE_SIZE
        self.calculate_viewport_size()


    def calculate_viewport_size(self):
        lat1, lon1 = Projector.num2deg(self.xtile, self.ytile, self.zoom)
        lat2, lon2 = Projector.num2deg(self.xtile + self.tiles_horizontally, self.ytile + self.tiles_vertically, self.zoom)
        self.viewport_w = utils.haversine(lat1=lat1, lon1=lon1, lat2=lat1, lon2=lon2)
        self.viewport_h = utils.haversine(lat1=lat1, lon1=lon1, lat2=lat2, lon2=lon1)


    def lonlat_to_screen(self, lon, lat):
        """
        Projects geodesic coordinates to screen
        :param lon: longitude
        :param lat: latitude
        :return: x,y screen coordinates
        """
        if type(lon) == list:
            lon = np.array(lon)
        if type(lat) == list:
            lat = np.array(lat)

        lat_rad = np.radians(lat)
        n = 2.0 ** self.zoom
        xtile = (lon + 180.0) / 360.0 * n
        ytile = (1.0 - np.log(np.tan(lat_rad) + (1 / np.cos(lat_rad))) / math.pi) / 2.0 * n
        x = (xtile * TILE_SIZE).astype(int)
        y = (SCREEN_H - ytile * TILE_SIZE).astype(int)
        return x, y


    def screen_to_latlon(self, x, y):
        """
        Return the latitude and longitude corresponding to a screen point
        :param x: screen x
        :param y: screen y
        :return: latitude and longitude at x,y
        """
        xtile = 1. * x / TILE_SIZE + self.xtile
        ytile = 1. * y / TILE_SIZE + self.ytile
        return self.num2deg(xtile, ytile, self.zoom)


class SetQueue(Queue):

    def _init(self, maxsize):
        self.queue = set()

    def _put(self, item):
        self.queue.add(item)

    def _get(self):
        return self.queue.pop()


class TileDownloaderThread(Thread):

    def __init__(self, queue):
        Thread.__init__(self)
        self.queue = queue
        self.daemon = True

    def run(self):
        while True:
            url, download_path = self.queue.get()
            assert download_path.endswith('.png')
            try:
                # print "downloading %s as %s" % (url, download_path)
                source = urllib2.urlopen(url)
                content = source.read()
                source.close()
                destination = open(download_path,'wb')
                destination.write(content)
                destination.close()
            except Exception as e:
                print(url, e)


_GEOPLOTLIB_ATTRIBUTION = u'made with geoplotlib | '

_DEFAULT_TILE_PROVIDERS = {
    'watercolor': { 'url': lambda zoom, xtile, ytile:
                            'http://%s.tile.stamen.com/watercolor/%d/%d/%d.png' % (random.choice(['a', 'b', 'c', 'd']), zoom, xtile, ytile),
                    'attribution': _GEOPLOTLIB_ATTRIBUTION + 'Map tiles by Stamen Design, under CC BY 3.0. Data by OpenStreetMap, under ODbL.'
    },
    'toner': { 'url': lambda zoom, xtile, ytile:
                        "http://%s.tile.stamen.com/toner/%d/%d/%d.png" % (random.choice(['a', 'b', 'c', 'd']), zoom, xtile, ytile),
               'attribution': _GEOPLOTLIB_ATTRIBUTION + 'Map tiles by Stamen Design, under CC BY 3.0. Data by OpenStreetMap, under ODbL.'
    },
    'toner-lite': { 'url': lambda zoom, xtile, ytile:
                            "http://%s.tile.stamen.com/toner-lite/%d/%d/%d.png" % (random.choice(['a', 'b', 'c', 'd']), zoom, xtile, ytile),
                    'attribution': _GEOPLOTLIB_ATTRIBUTION + 'Map tiles by Stamen Design, under CC BY 3.0. Data by OpenStreetMap, under ODbL.'
    },
    'darkmatter': { 'url': lambda zoom, xtile, ytile:
                            'http://%s.basemaps.cartocdn.com/dark_all/%d/%d/%d.png' % (random.choice(['a', 'b', 'c']), zoom, xtile, ytile),
                    'attribution': _GEOPLOTLIB_ATTRIBUTION + u'© OpenStreetMap contributors © CartoDB'
    },
    'positron': { 'url': lambda zoom, xtile, ytile:
                            'http://%s.basemaps.cartocdn.com/light_all/%d/%d/%d.png' % (random.choice(['a', 'b', 'c']), zoom, xtile, ytile),
                    'attribution': _GEOPLOTLIB_ATTRIBUTION + u'© OpenStreetMap contributors © CartoDB'
    }

}

class MapLayer():

    def __init__(self, tiles_provider, skipdl=False):
        if type(tiles_provider) == str:
            if tiles_provider in _DEFAULT_TILE_PROVIDERS:
                self.tiles_dir = tiles_provider
                self.url_generator = _DEFAULT_TILE_PROVIDERS[tiles_provider]['url']
                self.attribution = _DEFAULT_TILE_PROVIDERS[tiles_provider]['attribution']
            else:
                raise Exception('unknown style ' + tiles_provider)
        else:
            self.tiles_dir = tiles_provider['tiles_dir']
            self.url_generator = tiles_provider['url']
            self.attribution = tiles_provider['attribution']

        self.skipdl = skipdl
        self.tiles_cache = {}
        self.download_queue = SetQueue()
        self.download_threads = [TileDownloaderThread(self.download_queue) for i in range(2)]
        for t in self.download_threads:
            t.start()


    def get_tile(self, zoom, xtile, ytile):
        if xtile < 0 or ytile < 0 or xtile >= 2**zoom or ytile >= 2**zoom:
            return None

        tile_image = self.tiles_cache.get((zoom, xtile, ytile))
        if tile_image is not None:
            return tile_image

        url = self.url_generator(zoom, xtile, ytile)
        dir_path = expanduser('~') + '/geoplotlib_tiles/%s/%d/%d/' % (self.tiles_dir, zoom, xtile)
        download_path = dir_path + '%d.png' % ytile

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if not os.path.isfile(download_path):
            if self.skipdl:
                return None
            else:
                self.download_queue.put((url, download_path))
        else:
            try:
                tile_image = pyglet.image.load(download_path)
                tile_image.blit(2*SCREEN_W, 2*SCREEN_H, 0) # blit offscreen to check if valid
                self.tiles_cache[(zoom, xtile, ytile)] = pyglet.sprite.Sprite(tile_image)
                return self.tiles_cache[(zoom, xtile, ytile)]
            except Exception as exc:
                print(exc)
                assert download_path.endswith('.png')
                os.unlink(download_path)
                return None


    def draw(self, proj):
        for x in range(int(proj.xtile), int(proj.xtile + proj.tiles_horizontally + 1)):
            for y in range(int(proj.ytile), int(proj.ytile + proj.tiles_vertically + 1)):
                tilesurf = self.get_tile(proj.zoom, x, y)
                if tilesurf is not None:
                    try:
                        tilesurf.x = int((x - proj.xtile)*TILE_SIZE)
                        tilesurf.y = int(SCREEN_H - (y - proj.ytile + 1)*TILE_SIZE)
                        tilesurf.draw()
                    except Exception as e:
                        print('exception blitting', x, y, proj.zoom, e)



import base64
import traceback
import pyglet
from geoplotlib.core import GeoplotlibApp
from core import FONT_NAME


class AppConfig:

    def __init__(self):
        self.reset()


    def reset(self):
        self.layers = []
        self.bbox = None
        self.savefig = None
        self.tiles_provider = 'positron'
        self.smoothing = False
        self.map_alpha = 255
        screen =  pyglet.canvas.get_display().get_default_screen()
        self.screen_w = int(screen.width * .9)
        self.screen_h = int(screen.height * .9)
        self.requested_zoom = None


_global_config = AppConfig()


def _runapp(app_config):
    app = GeoplotlibApp(app_config)
    try:
        app.start()
    except:
        traceback.print_exc()
    finally:
        app.close()
        _global_config.reset()


def show():
    """Launch geoplotlib"""
    _runapp(_global_config)


def savefig(fname):
    """Launch geoplotlib, saves a screeshot and terminates"""
    _global_config.savefig = fname
    _runapp(_global_config)


def inline(width=900):
    """display the map inline in ipython

    :param width: image width for the browser
    """
    from IPython.display import Image, HTML, display, clear_output
    import random
    import string
    import urllib
    import os

    while True:
        fname = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(32))
        if not os.path.isfile(fname + '.png'):
            break

    savefig(fname)

    if os.path.isfile(fname + '.png'):
        with open(fname + '.png', 'rb') as fin:
            encoded = base64.b64encode(fin.read())
            b64 = urllib.parse.quote(encoded)

        image_html = "<img style='width: %dpx; margin: 0px; float: left; border: 1px solid black;' src='data:image/png;base64,%s' />" % (width, b64)

        display(HTML(image_html))
        os.remove(fname + '.png')


def dot(data, color=None, point_size=2, f_tooltip=None):
    """Create a dot density map

    :param data: data access object
    :param color: color
    :param point_size: point size
    :param f_tooltip: function to return a tooltip string for a point
    """
    from geoplotlib.layers import DotDensityLayer
    _global_config.layers.append(DotDensityLayer(data, color=color, point_size=point_size, f_tooltip=f_tooltip))


def scatter(data, color=None, point_size=2, f_tooltip=None):
    """Deprecated: use dot
    """
    import warnings
    warnings.warn("deprecated, use geoplotlib.dot", DeprecationWarning)
    dot(data, color, point_size, f_tooltip)


def hist(data, cmap='hot', alpha=220, colorscale='sqrt', binsize=16, show_tooltip=False,
         scalemin=0, scalemax=None, f_group=None, show_colorbar=True):
    """Create a 2D histogram

    :param data: data access object
    :param cmap: colormap name
    :param alpha: color alpha
    :param colorscale: scaling [lin, log, sqrt]
    :param binsize: size of the hist bins
    :param show_tooltip: if True, will show the value of bins on mouseover
    :param scalemin: min value for displaying a bin
    :param scalemax: max value for a bin
    :param f_group: function to apply to samples in the same bin. Default is to count
    :param show_colorbar: show colorbar
    """
    from geoplotlib.layers import HistogramLayer
    _global_config.layers.append(HistogramLayer(data, cmap=cmap, alpha=alpha, colorscale=colorscale,
                                    binsize=binsize, show_tooltip=show_tooltip, scalemin=scalemin, 
                                    scalemax=scalemax, f_group=f_group, show_colorbar=show_colorbar))


def graph(data, src_lat, src_lon, dest_lat, dest_lon, linewidth=1, alpha=220, color='hot',
          levels=10, color_by = None, seg_scale='log'):
    """Create a graph drawing a line between each pair of (src_lat, src_lon) and (dest_lat, dest_lon)

    :param data: data access object
    :param src_lat: field name of source latitude
    :param src_lon: field name of source longitude
    :param dest_lat: field name of destination latitude
    :param dest_lon: field name of destination longitude
    :param linewidth: line width
    :param alpha: color alpha
    :param color: color or colormap
    :param levels: coloring levels
    :param color_by: attribute name for color, default using node distance
    :param seg_scale: coloring data segamentation sacle, 'log' or 'lin',
        'lin' only used if not by distance
    """
    from geoplotlib.layers import GraphLayer
    _global_config.layers.append(GraphLayer(data, src_lat, src_lon, dest_lat, dest_lon, linewidth, alpha,
                                            color,levels, color_by, seg_scale))


def shapefiles(fname, f_tooltip=None, color=None, linewidth=3, shape_type='full', encoding='utf-8', encodingErrors='strict'):
    """
    Load and draws shapefiles

    :param fname: full path to the shapefile
    :param f_tooltip: function to generate a tooltip on mouseover
    :param color: color
    :param linewidth: line width
    :param shape_type: either full or bbox
    """
    from geoplotlib.layers import ShapefileLayer
    _global_config.layers.append(ShapefileLayer(fname, f_tooltip, color, linewidth, shape_type, encoding, encodingErrors))


def voronoi(data, line_color=None, line_width=2, f_tooltip=None, cmap=None, max_area=1e4, alpha=220):
    """
    Draw the voronoi tesselation of the points

    :param data: data access object
    :param line_color: line color
    :param line_width: line width
    :param f_tooltip: function to generate a tooltip on mouseover
    :param cmap: color map
    :param max_area: scaling constant to determine the color of the voronoi areas
    :param alpha: color alpha
    """
    from geoplotlib.layers import VoronoiLayer
    _global_config.layers.append(VoronoiLayer(data, line_color, line_width, f_tooltip, cmap, max_area, alpha))


def delaunay(data, line_color=None, line_width=2, cmap=None, max_lenght=100):
    """
    Draw a delaunay triangulation of the points

    :param data: data access object
    :param line_color: line color
    :param line_width: line width
    :param cmap: color map
    :param max_lenght: scaling constant for coloring the edges
    """
    from geoplotlib.layers import DelaunayLayer
    _global_config.layers.append(DelaunayLayer(data, line_color, line_width, cmap, max_lenght))


def convexhull(data, col, fill=True, point_size=4):
    """
    Convex hull for a set of points

    :param data: points
    :param col: color
    :param fill: whether to fill the convexhull polygon or not
    :param point_size: size of the points on the convexhull. Points are not rendered if None
    """
    from geoplotlib.layers import ConvexHullLayer
    _global_config.layers.append(ConvexHullLayer(data, col, fill, point_size))


def kde(data, bw, cmap='hot', method='hist', scaling='sqrt', alpha=220,
                 cut_below=None, clip_above=None, binsize=1, cmap_levels=10, show_colorbar=False):
    """
    Kernel density estimation visualization

    :param data: data access object
    :param bw: kernel bandwidth (in screen coordinates)
    :param cmap: colormap
    :param method: if kde use KDEMultivariate from statsmodel, which provides a more accurate but much slower estimation.
        If hist, estimates density applying gaussian smoothing on a 2D histogram, which is much faster but less accurate
    :param scaling: colorscale, lin log or sqrt
    :param alpha: color alpha
    :param cut_below: densities below cut_below are not drawn
    :param clip_above: defines the max value for the colorscale
    :param binsize: size of the bins for hist estimator
    :param cmap_levels: discretize colors into cmap_levels levels
    :param show_colorbar: show colorbar
    """
    from geoplotlib.layers import KDELayer
    _global_config.layers.append(KDELayer(data, bw, cmap, method, scaling, alpha,
                 cut_below, clip_above, binsize, cmap_levels, show_colorbar))


def markers(data, marker, f_tooltip=None, marker_preferred_size=32):
    """
    Draw markers

    :param data: data access object
    :param marker: full filename of the marker image
    :param f_tooltip: function to generate a tooltip on mouseover
    :param marker_preferred_size: size in pixel for the marker images
    """
    from geoplotlib.layers import MarkersLayer
    _global_config.layers.append(MarkersLayer(data, marker, f_tooltip, marker_preferred_size))


def geojson(filename, color='b', linewidth=1, fill=False, f_tooltip=None):
    """
    Draw features described in geojson format (http://geojson.org/)

    :param filename: filename of the geojson file
    :param color: color for the shapes. If callable, it will be invoked for each feature, passing the properties element
    :param linewidth: line width
    :param fill: if fill=True the feature polygon is filled, otherwise just the border is rendered
    :param f_tooltip: function to generate a tooltip on mouseover. It will be invoked for each feature, passing the properties element
    """
    from geoplotlib.layers import GeoJSONLayer
    _global_config.layers.append(GeoJSONLayer(filename, color=color, linewidth=linewidth, fill=fill, f_tooltip=f_tooltip))


def labels(data, label_column, color=None, font_name=FONT_NAME, 
           font_size=14, anchor_x='left', anchor_y='top'):
    """
    Draw a text label for each sample

    :param data: data access object
    :param label_column: column in the data access object where the labels text is stored
    :param color: color
    :param font_name: font name
    :param font_size: font size
    :param anchor_x: anchor x
    :param anchor_y: anchor y
    """
    from geoplotlib.layers import LabelsLayer
    _global_config.layers.append(LabelsLayer(data, label_column, color, font_name, 
           font_size, anchor_x, anchor_y))


def grid(lon_edges, lat_edges, values, cmap, alpha=255, vmin=None, vmax=None, levels=10, colormap_scale='lin', show_colorbar=True):
    """
        Values on a uniform grid
        
        :param lon_edges: longitude edges
        :param lat_edges: latitude edges
        :param values: matrix representing values on the grid 
        :param cmap: colormap name
        :param alpha: color alpha
        :param vmin: minimum value for the colormap
        :param vmax: maximum value for the colormap
        :param levels: number of levels for the colormap
        :param colormap_scale: colormap scale
        :param show_colorbar: show the colorbar in the UI
        """
    from geoplotlib.layers import GridLayer
    _global_config.layers.append(
        GridLayer(lon_edges, lat_edges, values, cmap, alpha, vmin, vmax, levels, colormap_scale, show_colorbar))


def clear():
    """
    Remove all existing layers
    """
    _global_config.layers = []


def tiles_provider(tiles_provider):
    """
    Set the tile provider

    :param tiles_provider: either one of the built-in providers
    ['watercolor', 'toner', 'toner-lite', 'mapquest', 'darkmatter','positron']
    or a custom provider in the form
    {'url': lambda zoom, xtile, ytile: 'someurl' % (zoom, xtile, ytile),
    'tiles_dir': 'mytiles',
    'attribution': 'my attribution'
    })
    """
    _global_config.tiles_provider = tiles_provider


def add_layer(layer):
    """
    Add a layer

    :param layer: a BaseLayer object
    """
    _global_config.layers.append(layer)


def set_bbox(bbox):
    """
    Set the map bounding box

    :param bbox: a BoundingBox object
    """
    _global_config.bbox = bbox


def set_smoothing(smoothing):
    """
    Enables OpenGL lines smoothing (antialiasing)

    :param smoothing: smoothing enabled or disabled
    """
    _global_config.smoothing = smoothing


def set_map_alpha(alpha):
    """
    Alpha color of the map tiles

    :param alpha: int between 0 and 255. 0 is completely dark, 255 is full brightness
    """
    if alpha < 0 or alpha > 255:
        raise Exception('invalid alpha '  + str(alpha))
    _global_config.map_alpha = alpha


def set_window_size(w, h):
    """
    Set the geoplotlib window size
    
    :param w: window width
    :param h: window height
    """
    _global_config.screen_w = w
    _global_config.screen_h = h


def request_zoom(zoom):
    _global_config.requested_zoom = zoom



from random import shuffle
import math


def _convert_color_format(col, alpha):
    return [int(c * 255) for c in col[:3]] + [alpha]


class ColorMap():

    def __init__(self, cmap_name, alpha=255, levels=10):
        """
        Converts continuous values into colors using matplotlib colorscales
        :param cmap_name: colormap name
        :param alpha: color alpha
        :param levels: discretize the colorscale into levels
        """
        from matplotlib.pylab import get_cmap
        self.cmap = get_cmap(cmap_name)
        self.alpha = alpha
        self.levels = levels
        self.mapping = {}


    def to_color(self, value, maxvalue, scale, minvalue=0.0):
        """
        convert continuous values into colors using matplotlib colorscales
        :param value: value to be converted
        :param maxvalue: max value in the colorscale
        :param scale: lin, log, sqrt
        :param minvalue: minimum of the input values in linear scale (default is 0)
        :return: the color corresponding to the value
        """
        
        if scale == 'lin':
            if minvalue >= maxvalue:
                raise Exception('minvalue must be less than maxvalue')
            else:
                value = 1.*(value-minvalue) / (maxvalue-minvalue)
        elif scale == 'log':
            if value < 1 or maxvalue <= 1:
                raise Exception('value and maxvalue must be >= 1')
            else:
                value = math.log(value) / math.log(maxvalue)
        elif scale == 'sqrt':
            if value < 0 or maxvalue <= 0:
                raise Exception('value and maxvalue must be greater than 0')
            else:
                value = math.sqrt(value) / math.sqrt(maxvalue)
        else:
            raise Exception('scale must be "lin", "log", or "sqrt"')

        if value < 0:
            value = 0
        elif value > 1:
            value = 1

        value = int(1.*self.levels*value)*1./(self.levels-1)

        if value not in self.mapping:
            self.mapping[value] = _convert_color_format(self.cmap(value), self.alpha)
        return self.mapping[value]


    def get_boundaries(self, maxvalue, scale):
        edges = []
        colors = []
        
        for i in range(self.levels+1):
            z = 1. * i / self.levels
            if scale == 'lin':
                edges.append(maxvalue * z)
            elif scale == 'log':
                # log(v)/log(maxvalue) = z
                # log(v) = log(maxvalue)*z
                # v = e^(log(maxvalue)*z)
                edges.append(math.exp(math.log(maxvalue)*z))
            elif scale == 'sqrt':
                # sqrt(v)/sqrt(maxvalue) = z
                # v/maxvalue = z^2
                # v = maxvalue * z^2
                edges.append(maxvalue * z**2)
            else:
                raise Exception('scale must be "lin", "log", or "sqrt"') 

        for e in edges[:-1]:
            colors.append(self.to_color(e, maxvalue, scale))

        return edges, colors


def create_set_cmap(values, cmap_name, alpha=255):
    """
    return a dict of colors corresponding to the unique values
    
    :param values: values to be mapped
    :param cmap_name: colormap name
    :param alpha: color alpha
    :return: dict of colors corresponding to the unique values
    """
    unique_values = list(set(values))
    shuffle(unique_values)
    from pylab import get_cmap
    cmap = get_cmap(cmap_name)
    d = {}
    for i in range(len(unique_values)):
        d[unique_values[i]] = _convert_color_format(cmap(1.*i/len(unique_values)), alpha)
    return d


def colorbrewer(values, alpha=255):
    """
    Return a dict of colors for the unique values.
    Colors are adapted from Harrower, Mark, and Cynthia A. Brewer.
    "ColorBrewer. org: an online tool for selecting colour schemes for maps."
    The Cartographic Journal 40.1 (2003): 27-37.

    :param values: values
    :param alpha: color alphs
    :return: dict of colors for the unique values.
    """
    basecolors = [
        [31, 120, 180],
        [178, 223, 138],
        [51, 160, 44],
        [251, 154, 153],
        [227, 26, 28],
        [253, 191, 111],
        [255, 127, 0],
        [202, 178, 214],
        [106, 61, 154],
        [255, 255, 153],
        [177, 89, 40]
    ]
    unique_values = list(set(values))
    return {k: basecolors[i % len(basecolors)] + [alpha]  for i, k in enumerate(unique_values)}



from collections import defaultdict
import csv
from datetime import datetime
import json
from math import radians, cos, sin, asin, sqrt
import urllib2
import numpy as np


def haversine(lon1, lat1, lon2, lat2):
    """
    Distance between geodesic coordinates http://www.movable-type.co.uk/scripts/latlong.html

    :param lon1: point 1 latitude
    :param lat1: point 1 longitude
    :param lon2: point 1 latitude
    :param lat2: point 2 longitude
    :return: distance in meters between points 1 and 2
    """
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))
    m = 6367000 * c
    return m


class DataAccessObject():
    """
    This class wraps data into a dict-like object
    """

    def __init__(self, dict_or_df):
        """
        Create a DataAccessObject either from a dictionary or a pandas.DataFrame
        """
        if type(dict_or_df) == dict:
            self.dict = dict_or_df
        else:
            from pandas import DataFrame
            if type(dict_or_df) == DataFrame:
                self.dict = {col: dict_or_df[col].values for col in dict_or_df.columns}
            else:
                raise Exception('dict_or_df must be either a dictionary or a pandas.DataFrame')


    @staticmethod
    def from_dataframe(df):
        """
        Loads data from a pandas DataFrame

        :param df: dataframe
        :return: a DataAccessObject
        """
        import warnings
        warnings.warn('use ctor directly instead', DeprecationWarning)
        return DataAccessObject(df)


    def __getitem__(self, key):
        return self.dict[key]


    def __setitem__(self, key, value):
        assert type(value) == np.ndarray
        self.dict[key] = value


    def __delitem__(self, key):
        del self.dict[key]


    def rename(self, mapping):
        """
        Rename fields

        :param mapping: a dict in the format {'oldkey1': 'newkey1', ...}
        """
        for old_key, new_key in mapping:
            self.dict[new_key] = self.dict[old_key]
            del self.dict[old_key]


    def where(self, mask):
        """
        :param mask: boolean mask
        :return: a DataAccessObject with a subset of rows matching mask
        """
        assert len(mask) == len(self)
        return DataAccessObject({k: self.dict[k][mask] for k in self.dict})


    def groupby(self, field1, field2=None):
        if field2 is None:
            uniquevalues = list(set(self.dict[field1]))
            return [(v, self.where(self.dict[field1] == v)) for v in uniquevalues]
        else:
            uniquevalues = set([tuple(row) for row in np.vstack([self.dict[field1],self.dict[field2]]).T])
            return [((v1,v2), self.where((self.dict[field1] == v1) & (self.dict[field2] == v2))) \
                    for v1,v2 in uniquevalues]


    def head(self, n):
        """
        Return a DataAccessObject containing the first n rows

        :param n: number of rows
        :return: DataAccessObject
        """
        return DataAccessObject({k: self.dict[k][:n] for k in self.dict})


    def keys(self):
        """
        :return: the keys (field names)
        """
        return self.dict.keys()


    def values(self):
        """
        :return: the values (field values)
        """
        return self.dict.values()


    def __str__(self):
        return 'DataAccessObject(%s x %d)' % (str(self.dict.keys()), len(self))


    def __repr__(self):
        return self.__str__()


    def __len__(self):
        return len(self.dict.values()[0])



def read_csv(fname):
    """
    Read a csv file into a DataAccessObject

    :param fname: filename
    """
    values = defaultdict(list)
    with open(fname) as f:
        reader = csv.DictReader(f)
        for row in reader:
            for (k,v) in row.items():
                values[k].append(v)
    npvalues = {k: np.array(values[k]) for k in values.keys()}
    for k in npvalues.keys():
        for datatype in [np.int, np.float]:
            try:
                npvalues[k][:1].astype(datatype)
                npvalues[k] = npvalues[k].astype(datatype)
                break
            except:
                pass
    dao = DataAccessObject(npvalues)
    return dao


def epoch_to_str(epoch, fmt='%Y-%m-%d %H:%M:%S'):
    """
    Convert a unix timestamp into date string

    :param epoch: unix timestamp
    :param fmt: date format
    :return: formatted date from timestamp
    """
    return datetime.fromtimestamp(epoch).strftime(fmt)


def parse_raw_str(v):
    try:
        v = v.decode('utf-8')
    except:
        try:
            v = v.decode('latin1')
        except:
            pass
    return v


class BoundingBox():

    def __init__(self, north, west, south, east):
        """
        Represent a map boundingbox

        :param north: northmost latitude
        :param west: westmost longitude
        :param south: southmost latitude
        :param east: eastmost longitude
        :return:
        """
        self.north = north
        self.west = west
        self.south = south
        self.east = east


    @staticmethod
    def from_points(lons, lats):
        """
        Compute the BoundingBox from a set of latitudes and longitudes

        :param lons: longitudes
        :param lats: latitudes
        :return: BoundingBox
        """
        north, west = max(lats), min(lons)
        south, east = min(lats), max(lons)
        return BoundingBox(north=north, west=west, south=south, east=east)


    @staticmethod
    def from_bboxes(bboxes):
        """
        Compute a BoundingBox enclosing all specified bboxes

        :param bboxes: a list of BoundingBoxes
        :return: BoundingBox
        """
        north = max([b.north for b in bboxes])
        south = min([b.south for b in bboxes])
        west = min([b.west for b in bboxes])
        east = max([b.east for b in bboxes])
        return BoundingBox(north=north, west=west, south=south, east=east)


    def __str__(self):
        return 'BoundingBox(north=%.6f, west=%.6f, south=%.6f, east=%.6f)' % (self.north, self.west, self.south, self.east)


    @staticmethod
    def from_nominatim(query):
        url = urllib2.urlopen('http://nominatim.openstreetmap.org/search.php?q=%s&format=json' % query)
        jo = json.load(url)

        if len(jo) == 0:
            raise Exception('No results found')

        south, north, west, east = map(float, jo[0]['boundingbox'])
        print('bbox from Nominatim:', south, north, west, east)
        return BoundingBox(north=north, west=west, south=south, east=east)


BoundingBox.WORLD = BoundingBox(north=85, west=-170, south=-85, east=190)
BoundingBox.DK = BoundingBox(north=57.769, west=7.932, south=54.444, east=13.282)
BoundingBox.DTU = BoundingBox(north=55.7925, west=12.5092, south=55.7784, east=12.5309)
BoundingBox.KBH = BoundingBox(north=55.8190, west=12.0369, south=55.5582, east=12.7002)
BoundingBox.DOWNTOWN = BoundingBox(north=55.728229, west=12.420230, south=55.629118, east=12.683902)
BoundingBox.USA = BoundingBox(north=51.338994, west=-124.349040, south=14.851581, east=-56.849040)


from setuptools import setup

setup(name='geoplotlibfixed',
      use_2to3=True,
      packages=['geoplotlibfixed'],
      version='0.0.1',
      description='python toolbox for geographic visualizations',
      author='John Landa',
      author_email='jonathanlanda@gmail.com',
      url='https://github.com/johnlanda/geoplotlib',
      download_url='https://github.com/johnlanda/geoplotlib-fixed/archive/v0.0.1.tar.gz',
      install_requires=['numpy>=1.12','pyglet>=1.2.4']
)
